{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with the
License. You may obtain a copy of the License at
http://www.mozilla.org/NPL/NPL-1_1Final.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: mwKeyProcessor.pas, released March 16, 2001.

The Initial Developer of the Original Code is Martin Waldenburg
(Martin.Waldenburg@T-Online.de).
Portions created by Martin Waldenburg are Copyright (C) 2000, 2001 Martin Waldenburg.
All Rights Reserved.

Contributor(s): ___________________.

Last Modified: 3/16/2001
Current Version: 1.0

Notes: This is a generic processor of key commands.

Known Issues:
-----------------------------------------------------------------------------}
{$R-}
unit mwKeyProcessor;

interface

uses Windows, Sysutils, Classes, Dialogs, Forms;

type
  MethodProc = procedure of Object;

  TmwKeyProcedure = class(TObject)
  private
    fChain: TList;
  public
    ToExecute: MethodProc;
    Key: String;
    constructor Create(const Method: MethodProc; const Str: String; const List: TList);
    destructor Destroy; override;
  end;

  PExecutePointerList = ^TExecutePointerList;
  TExecutePointerList = array[1..1] of TObject;

  TmwBaseKeyProcessor = class(TObject)
    FArray: PExecutePointerList;
    fCapacity: Integer;
  protected
    function Get(Index: Integer): Pointer;
    procedure Put(Index: Integer; Item: Pointer);
    procedure SetCapacity(NewCapacity: Integer);
    property Capacity: Integer read fCapacity;
    property Items[Index: Integer]: Pointer read Get write Put; default;
  public
    destructor Destroy; override;
    procedure Clear;
  end;

  TmwKeyProcessorStrings = class(TmwBaseKeyProcessor)
  protected
    procedure Add(const aKeyProcedure: TmwKeyProcedure);
  end;

  TmwKeyProcessorItems = class(TmwBaseKeyProcessor)
  protected
    procedure Add(const aKeyProcedure: TmwKeyProcedure);
    function Execute(const KeyString: string): Boolean;
    function Hash(const aString: string): Integer;
  end;

  TmwKeyProcessor = class(TmwBaseKeyProcessor)
  private
    fObject: TObject;
    fChain: TList;
    FKeyString: String;
    FShiftState: TShiftState;
    fLastKey: Word;
    function GetCount: Integer;
    function GetEntry(Index: Integer): TmwKeyProcedure;
  protected
    procedure Add(const aKeyProcedure: TmwKeyProcedure);
    procedure AddToKeyString(Key: Word);
    procedure ClearKeyString;
    function Execute(const KeyString: string): Boolean;
    function Hash(const aString: string): Integer;
    function MethodByName(Name: ShortString): MethodProc;
    function MethodToName(Method: MethodProc): ShortString;
    property Chain: TList read fChain;
    property KeyString: String read FKeyString write FKeyString;
    property ShiftState: TShiftState read FShiftState write FShiftState;
  public
    constructor Create(AnObject: TObject);
    destructor Destroy; override;
    procedure AddKey(const Method: MethodProc; const KeyString: String);
    procedure KeyDown(var Key: Word; Shift: TShiftState); dynamic;
    procedure KeyUp(var Key: Word; Shift: TShiftState); dynamic;
    procedure KeyPress(var Key: Char); dynamic;
    procedure LoadFromFile(FileName: String);
    procedure SaveToFile(FileName: String);
    property Count: Integer read GetCount;
    property Entry[Index: Integer]: TmwKeyProcedure read GetEntry;
  end;

procedure InitTables;

implementation

var
  mwInsensitiveHashTable: array[#0..#255] of Byte;

procedure InitTables;
var
  I, K: Char;
  Temp: Byte;
begin
  for I := #0 to #255 do
  begin
    mwInsensitiveHashTable[I] := Ord(I);
  end;
  RandSeed := 255;
  for I := #1 to #255 do
  begin
    repeat
      K := Char(Random(255));
    until K <> #0;
    Temp := mwInsensitiveHashTable[I];
    mwInsensitiveHashTable[I] := mwInsensitiveHashTable[K];
    mwInsensitiveHashTable[K] := Temp;
  end;
  for I := #0 to #255 do
    mwInsensitiveHashTable[I] := mwInsensitiveHashTable[AnsiLowerCase(string(I))[1]];
end;

function KeyToStr(Key: Integer): String;
begin
  case Key of
    VK_LBUTTON: Result:= 'LBUTTON';
    VK_RBUTTON: Result:= 'RBUTTON';
    VK_CANCEL: Result:= 'CANCEL';
    VK_MBUTTON: Result:= 'MBUTTON';
    VK_BACK: Result:= 'BACK';
    VK_TAB: Result:= 'TAB';
    VK_CLEAR: Result:= 'CLEAR';
    VK_RETURN: Result:= 'RETURN';
    VK_SHIFT: Result:= 'SHIFT';
    VK_CONTROL: Result:= 'CONTROL'; 
    VK_MENU: Result:= 'MENU';
    VK_PAUSE: Result:= 'PAUSE';
    VK_CAPITAL: Result:= 'CAPITAL';
    VK_KANA: Result:= 'KANA';
    VK_JUNJA: Result:= 'JUNJA';
    VK_FINAL: Result:= 'FINAL';
    VK_HANJA: Result:= 'HANJA';
    VK_CONVERT: Result:= 'CONVERT';
    VK_NONCONVERT: Result:= 'NONCONVERT';
    VK_ACCEPT: Result:= 'ACCEPT';
    VK_MODECHANGE: Result:= 'MODECHANGE';
    VK_ESCAPE: Result:= 'ESCAPE';
    VK_SPACE: Result:= 'SPACE';
    VK_PRIOR: Result:= 'PRIOR';
    VK_NEXT: Result:= 'NEXT';
    VK_END: Result:= 'END';
    VK_HOME: Result:= 'HOME';
    VK_LEFT: Result:= 'LEFT';
    VK_UP: Result:= 'UP';
    VK_RIGHT: Result:= 'RIGHT';
    VK_DOWN: Result:= 'DOWN';
    VK_SELECT: Result:= 'SELECT';
    VK_PRINT: Result:= 'PRINT';
    VK_EXECUTE: Result:= 'EXECUTE';
    VK_SNAPSHOT: Result:= 'SNAPSHOT';
    VK_INSERT: Result:= 'INSERT';
    VK_DELETE: Result:= 'DELETE';
    VK_HELP: Result:= 'HELP';
    VK_LWIN: Result:= 'LWIN';
    VK_RWIN: Result:= 'RWIN';
    VK_APPS: Result:= 'APPS';
    VK_NUMPAD0: Result:= 'NUMPAD0';
    VK_NUMPAD1: Result:= 'NUMPAD1';
    VK_NUMPAD2: Result:= 'NUMPAD2';
    VK_NUMPAD3: Result:= 'NUMPAD3';
    VK_NUMPAD4: Result:= 'NUMPAD4';
    VK_NUMPAD5: Result:= 'NUMPAD5';
    VK_NUMPAD6: Result:= 'NUMPAD6';
    VK_NUMPAD7: Result:= 'NUMPAD7';
    VK_NUMPAD8: Result:= 'NUMPAD8';
    VK_NUMPAD9: Result:= 'NUMPAD9';
    VK_MULTIPLY: Result:= 'MULTIPLY';
    VK_ADD: Result:= 'ADD'; 
    VK_SEPARATOR: Result:= 'SEPARATOR';
    VK_SUBTRACT: Result:= 'SUBTRACT';
    VK_DECIMAL: Result:= 'DECIMAL';
    VK_DIVIDE: Result:= 'DIVIDE'; 
    VK_F1: Result:= 'F1';
    VK_F2: Result:= 'F2'; 
    VK_F3: Result:= 'F3';
    VK_F4: Result:= 'F4'; 
    VK_F5: Result:= 'F5';
    VK_F6: Result:= 'F6';
    VK_F7: Result:= 'F7';
    VK_F8: Result:= 'F8'; 
    VK_F9: Result:= 'F9';
    VK_F10: Result:= 'F10';
    VK_F11: Result:= 'F11';
    VK_F12: Result:= 'F12'; 
    VK_F13: Result:= 'F13';
    VK_F14: Result:= 'F14'; 
    VK_F15: Result:= 'F15';
    VK_F16: Result:= 'F16'; 
    VK_F17: Result:= 'F17';
    VK_F18: Result:= 'F18'; 
    VK_F19: Result:= 'F19';
    VK_F20: Result:= 'F20'; 
    VK_F21: Result:= 'F21';
    VK_F22: Result:= 'F22'; 
    VK_F23: Result:= 'F23';
    VK_F24: Result:= 'F24'; 
    VK_NUMLOCK: Result:= 'NUMLOCK';
    VK_SCROLL: Result:= 'SCROLL'; 
    VK_LSHIFT: Result:= 'LSHIFT';
    VK_RSHIFT: Result:= 'RSHIFT'; 
    VK_LCONTROL: Result:= 'LCONTROL';
    VK_RCONTROL: Result:= 'RCONTROL'; 
    VK_LMENU: Result:= 'LMENU';
    VK_RMENU: Result:= 'RMENU'; 
    VK_PROCESSKEY: Result:= 'PROCESSKEY';
    VK_ATTN: Result:= 'ATTN'; 
    VK_CRSEL: Result:= 'CRSEL';
    VK_EXSEL: Result:= 'EXSEL'; 
    VK_EREOF: Result:= 'EREOF';
    VK_PLAY: Result:= 'PLAY';
    VK_ZOOM: Result:= 'ZOOM';
    VK_NONAME: Result:= 'NONAME'; 
    VK_PA1: Result:= 'PA1';
    VK_OEM_CLEAR: Result:= 'OEM_CLEAR';
  else Result:= Char(Key);
  end;
end;

{ TmwKeyProcedure }

constructor TmwKeyProcedure.Create(const Method: MethodProc; const Str: String; const List: TList);
begin
  ToExecute:= Method;
  Key:= Str;
  fChain:= List;
  fChain.Add(Self);
end;

destructor TmwKeyProcedure.Destroy;
begin
  fChain.Remove(Self);
  inherited Destroy;
end;

{ TmwBaseKeyProcessor }

procedure TmwBaseKeyProcessor.Clear;
var
  I: Integer;
begin
  for I := 1 to fCapacity do
    if FArray[I] <> nil then
      FArray[I].Free;
  ReallocMem(FArray, 0);
  fCapacity := 0;
end;

destructor TmwBaseKeyProcessor.Destroy;
begin
  Clear;
  inherited Destroy;
end;

function TmwBaseKeyProcessor.Get(Index: Integer): Pointer;
begin
  Result := nil;
  if (Index > 0) and (Index <= fCapacity) then
    Result := FArray[Index];
end;

procedure TmwBaseKeyProcessor.Put(Index: Integer; Item: Pointer);
begin
  if (Index > 0) and (Index <= fCapacity) then
    FArray[Index] := Item;
end;

procedure TmwBaseKeyProcessor.SetCapacity(NewCapacity: Integer);
var
  I, OldCapacity: Integer;
begin
  if NewCapacity > fCapacity then
  begin
    ReallocMem(FArray, (NewCapacity) * SizeOf(Pointer));
    OldCapacity := fCapacity;
    FCapacity := NewCapacity;
    for I := OldCapacity + 1 to NewCapacity do Items[I] := nil;
  end;
end;

{ TmwKeyProcessorStrings }

procedure TmwKeyProcessorStrings.Add(const aKeyProcedure: TmwKeyProcedure);
begin
  SetCapacity(Capacity + 1);
  FArray[Capacity] := aKeyProcedure;
end;

{ TmwKeyProcessorItems }

procedure TmwKeyProcessorItems.Add(const aKeyProcedure: TmwKeyProcedure);
var
  KeyProcedure: TmwKeyProcedure;
  ExecuteStrings: TmwKeyProcessorStrings;
  ExecuteVal: Integer;
begin
  ExecuteVal := Hash(aKeyProcedure.Key);
  SetCapacity(ExecuteVal);
  if Items[ExecuteVal] = nil then
  begin
    Items[ExecuteVal] := aKeyProcedure;
  end else
    if FArray[ExecuteVal] is TmwKeyProcessorStrings then
    begin
      TmwKeyProcessorStrings(Items[ExecuteVal]).Add(aKeyProcedure);
    end else
    begin
      KeyProcedure := Items[ExecuteVal];
      ExecuteStrings := TmwKeyProcessorStrings.Create;
      Items[ExecuteVal] := ExecuteStrings;
      ExecuteStrings.Add(KeyProcedure);
      ExecuteStrings.Add(aKeyProcedure)
    end;
end;

function TmwKeyProcessorItems.Execute(const KeyString: string): Boolean;
var
  I: Integer;
  KeyProc: TmwKeyProcedure;
  HashValue: Integer;
begin
  Result := False;
  HashValue:= Hash(KeyString);
  if (HashValue < 1) or (HashValue > Capacity) then Exit;
  if Items[HashValue] <> nil then
  begin
    if FArray[HashValue] is TmwKeyProcedure then
    begin
      if SameText(TmwKeyProcedure(FArray[HashValue]).Key, KeyString) then
      begin
        Result := True;
        KeyProc:= TmwKeyProcedure(FArray[HashValue]);
        KeyProc.ToExecute;
      end;
    end else
      for I := 1 to TmwKeyProcessorStrings(FArray[HashValue]).Capacity do
      begin
        if SameText(TmwKeyProcedure(TmwKeyProcessorStrings(FArray[HashValue])[I]).Key, KeyString) then
        begin
          Result := True;
          KeyProc := TmwKeyProcedure(TmwKeyProcessorStrings(FArray[HashValue])[I]);
          KeyProc.ToExecute;
          exit;
        end;
      end;
  end;
end;

function TmwKeyProcessorItems.Hash(const aString: string): Integer;
var
  I: Integer;
begin
  Result:= 0;
  for I:= 1 to Length(aString) do
  begin
    inc(Result, mwInsensitiveHashTable[aString[I]]);
    Result:= Result mod 7 +1;
  end;
end;

{ TmwKeyProcessor }

procedure TmwKeyProcessor.Add(const aKeyProcedure: TmwKeyProcedure);
var
  KeyProcedure: TmwKeyProcedure;
  HashValue: Integer;
  KeyProcessorItems: TmwKeyProcessorItems;
begin
  HashValue := Hash(aKeyProcedure.Key);
  SetCapacity(HashValue);
  if Items[HashValue] = nil then
  begin
    Items[HashValue] := aKeyProcedure;
  end else
    if FArray[HashValue] is TmwKeyProcessorItems then
    begin
      TmwKeyProcessorItems(Items[HashValue]).Add(aKeyProcedure);
    end else
    begin
      KeyProcedure := Items[HashValue];
      KeyProcessorItems := TmwKeyProcessorItems.Create;
      Items[HashValue] := KeyProcessorItems;
      KeyProcessorItems.Add(KeyProcedure);
      KeyProcessorItems.Add(aKeyProcedure);
    end;
end;

procedure TmwKeyProcessor.AddKey(const Method: MethodProc; const KeyString: String);
begin
  Add(TmwKeyProcedure.Create(Method, KeyString, fChain));
end;

procedure TmwKeyProcessor.AddToKeyString(Key: Word);
var
  Temp: String;
begin
  Temp:= KeyToStr(Key);
  if (Key in [VK_Menu, VK_CONTROL, VK_SHIFT]) and (Key = FLastKey) then Exit;
  if ((Key = VK_Control) and (FLastKey = VK_Menu)) then ClearKeyString;
  FKeyString:= FKeyString + '*' + Temp;
  FLastKey:= Key;
end;

procedure TmwKeyProcessor.ClearKeyString;
begin
  fKeyString:= '';
end;

constructor TmwKeyProcessor.Create(AnObject: TObject);
begin
  inherited Create;
  fObject:= AnObject;
  fChain:= TList.Create;
end;

destructor TmwKeyProcessor.Destroy;
begin
  fChain.Free;
  inherited Destroy;
end;

function TmwKeyProcessor.Execute(const KeyString: string): Boolean;
var
  HashValue: Integer;
begin
  Result := False;
  HashValue:= Hash(KeyString);
  if (HashValue < 1) or (HashValue > Capacity) then Exit;
  if Items[HashValue] <> nil then
  begin
    if FArray[HashValue] is TmwKeyProcedure then
    begin
      if SameText(TmwKeyProcedure(Items[HashValue]).Key, KeyString) then
      begin
        Result:= True;
        TmwKeyProcedure(Items[HashValue]).ToExecute;
      end;
    end else
    begin
      Result:= TmwKeyProcessorItems(FArray[HashValue]).Execute(KeyString);
    end;
  end;
end;

function TmwKeyProcessor.GetCount: Integer;
begin
  Result:= fChain.Count;
end;

function TmwKeyProcessor.GetEntry(Index: Integer): TmwKeyProcedure;
begin
  Result:= fChain[Index];
end;

function TmwKeyProcessor.Hash(const aString: string): Integer;
var
  I: Integer;
begin
  Result:= 0;
  for I:= 1 to Length(aString) do
  begin
    inc(Result, mwInsensitiveHashTable[aString[I]]);
    Result:= Result mod 61 +1;
  end;
end;

procedure TmwKeyProcessor.KeyDown(var Key: Word; Shift: TShiftState);
begin
  if Shift =[] then ClearKeyString;
  fShiftState:= Shift;
  Case Key of
    VK_Menu, VK_CONTROL, VK_SHIFT: AddToKeyString(Key);
  else
    begin
      AddToKeyString(Key);
      if Execute(KeyString) then ClearKeyString;
    end;
  end;
end;

procedure TmwKeyProcessor.KeyPress(var Key: Char);
begin
  if fShiftState = [] then ClearKeyString;
end;

procedure TmwKeyProcessor.KeyUp(var Key: Word; Shift: TShiftState);
begin
  Case Key of
    VK_Menu: Exclude(fShiftState, ssAlt);
    VK_CONTROL: Exclude(fShiftState, ssCtrl);
    VK_SHIFT: Exclude(fShiftState, ssShift);
  end;;
end;

procedure TmwKeyProcessor.LoadFromFile(FileName: String);
var
  List: TStringList;
  I: Integer;
  Position: Word;
  KeyString, Name, Temp: String;
begin
  Clear;
  List := TStringList.Create;
  List.LoadFromFile(FileName);
  for I := 0 to List.Count - 1 do
  begin
    Temp:= List[I];
    Position:= pos('*', Temp);
    Name:= Copy(Temp, 1, Position -1);
    KeyString:= Copy(Temp, Position, MaxInt);
    AddKey(MethodByName(Name), KeyString);
  end;
  List.Free;
end;

function TmwKeyProcessor.MethodByName(Name: ShortString): MethodProc;
var
  Ptr: Pointer;
  function PointerToMethod(const C): MethodProc;
  begin
    Result := MethodProc(C);
  end;
begin
  try
    Ptr:= fObject.MethodAddress(Name);
    Result:= PointerToMethod(Ptr);
  except
  end;
end;

function TmwKeyProcessor.MethodToName(Method: MethodProc): ShortString;
begin
  try
    Result := FObject.MethodName(@Method);
  except
  end;
end;

procedure TmwKeyProcessor.SaveToFile(FileName: String);
var
  List: TStringList;
  I: Integer;
  Item: TmwKeyProcedure;
begin
  List := TStringList.Create;
  for I := 0 to fChain.Count - 1 do
  begin
    Item:= TmwKeyProcedure(fChain[I]);
    List.Add(MethodToName(Item.ToExecute) + Item.Key);
  end;
  List.SaveToFile(FileName);
  List.Free;
end;

{$R+}
initialization
  InitTables;
end.

